@@ -26,133 +26,8 @@ incomplete.
 .. _CFFI: http://cffi.readthedocs.org/
 """
 
-from snimpy import cffi_fix  # nopep8
-from cffi import FFI
-
-ffi = FFI()
-ffi.cdef("""
-
-typedef char *SmiIdentifier;
-typedef unsigned long SmiUnsigned32;
-typedef long SmiInteger32;
-typedef unsigned long long SmiUnsigned64;
-typedef long long SmiInteger64;
-typedef unsigned int SmiSubid;
-typedef float SmiFloat32;
-typedef double SmiFloat64;
-typedef long double SmiFloat128;
-
-typedef enum SmiBasetype {
-    SMI_BASETYPE_INTEGER32,
-    SMI_BASETYPE_OCTETSTRING,
-    SMI_BASETYPE_OBJECTIDENTIFIER,
-    SMI_BASETYPE_UNSIGNED32,
-    SMI_BASETYPE_INTEGER64,
-    SMI_BASETYPE_UNSIGNED64,
-    SMI_BASETYPE_ENUM,
-    SMI_BASETYPE_BITS,
-    ...
-} SmiBasetype;
-typedef struct SmiType {
-    SmiIdentifier       name;
-    SmiBasetype         basetype;
-    char                *format;
-    ...;
-} SmiType;
-
-typedef enum SmiIndexkind {
-    SMI_INDEX_INDEX,
-    SMI_INDEX_AUGMENT,
-    ...
-} SmiIndexkind;
-
-typedef unsigned int SmiNodekind;
-#define SMI_NODEKIND_NODE         ...
-#define SMI_NODEKIND_SCALAR       ...
-#define SMI_NODEKIND_TABLE        ...
-#define SMI_NODEKIND_ROW          ...
-#define SMI_NODEKIND_COLUMN       ...
-
-typedef struct SmiNode {
-    SmiIdentifier       name;
-    unsigned int        oidlen;
-    SmiSubid            *oid;
-    char                *format;
-    SmiIndexkind        indexkind;
-    int                 implied;
-    SmiNodekind         nodekind;
-    ...;
-} SmiNode;
-
-typedef struct SmiValue {
-    SmiBasetype             basetype;
-    union {
-        SmiUnsigned64       unsigned64;
-        SmiInteger64        integer64;
-        SmiUnsigned32       unsigned32;
-        SmiInteger32        integer32;
-        SmiFloat32          float32;
-        SmiFloat64          float64;
-        SmiFloat128         float128;
-        SmiSubid            *oid;
-        char                *ptr;
-    } value;
-    ...;
-} SmiValue;
-
-typedef struct SmiRange {
-    SmiValue            minValue;
-    SmiValue            maxValue;
-} SmiRange;
-
-typedef struct SmiModule {
-    SmiIdentifier       name;
-    int                 conformance;
-    ...;
-} SmiModule;
-
-typedef struct SmiElement {
-    ...;
-} SmiElement;
-
-typedef struct SmiNamedNumber {
-    SmiIdentifier       name;
-    SmiValue            value;
-} SmiNamedNumber;
-
-int          smiInit(const char *);
-void         smiExit(void);
-void         smiSetErrorLevel(int);
-void         smiSetFlags(int);
-char        *smiLoadModule(const char *);
-SmiModule   *smiGetModule(const char *);
-SmiModule   *smiGetNodeModule(SmiNode *);
-SmiType     *smiGetNodeType(SmiNode *);
-SmiType     *smiGetParentType(SmiType *);
-char        *smiRenderNode(SmiNode *, int);
-SmiElement  *smiGetFirstElement(SmiNode *);
-SmiElement  *smiGetNextElement(SmiElement *);
-SmiNode     *smiGetElementNode(SmiElement *);
-SmiRange    *smiGetFirstRange(SmiType *);
-SmiRange    *smiGetNextRange(SmiRange *);
-SmiNode     *smiGetNode(SmiModule *, const char *);
-SmiNode     *smiGetFirstNode(SmiModule *, SmiNodekind);
-SmiNode     *smiGetNextNode(SmiNode *, SmiNodekind);
-SmiNode     *smiGetParentNode(SmiNode *);
-SmiNode     *smiGetRelatedNode(SmiNode *);
-SmiNode     *smiGetFirstChildNode(SmiNode *);
-SmiNode     *smiGetNextChildNode(SmiNode *);
-SmiNamedNumber *smiGetFirstNamedNumber(SmiType *);
-SmiNamedNumber *smiGetNextNamedNumber(SmiNamedNumber *);
-
-#define SMI_FLAG_ERRORS ...
-#define SMI_FLAG_RECURSIVE ...
-#define SMI_RENDER_ALL ...
-""")
-
-_smi = ffi.verify("""
-#include <smi.h>
-""", libraries=["smi"])
+from snimpy.smi import smi as _smi
+from snimpy.smi import ffi
 
 
 class SMIException(Exception):
@@ -174,6 +49,7 @@ class Node(object):
         :param node: libsmi node supporting this node.
         """
         self.node = node
+        self._override_type = None
 
     @property
     def type(self):
@@ -185,7 +61,10 @@ class Node(object):
             an appropriate representation.
         """
         from snimpy import basictypes
-        t = _smi.smiGetNodeType(self.node)
+        if self._override_type:
+            t = self._override_type
+        else:
+            t = _smi.smiGetNodeType(self.node)
         target = {
             _smi.SMI_BASETYPE_INTEGER32: basictypes.Integer,
             _smi.SMI_BASETYPE_INTEGER64: basictypes.Integer,
@@ -210,6 +89,63 @@ class Node(object):
             raise SMIException("unable to retrieve type of node")
         return target
 
+    @property
+    def typeName(self):
+        """Retrieves the name of the the node's current declared type
+        (not basic type).
+
+        :return: A string representing the current declared type,
+            suitable for assignment to type.setter.
+        """
+        if self._override_type:
+            t = self._override_type
+        else:
+            t = _smi.smiGetNodeType(self.node)
+
+        # This occurs when the type is "implied".
+        if t.name == ffi.NULL:
+            t = _smi.smiGetParentType(t)
+
+        if t is None or t == ffi.NULL:
+            raise SMIException("unable to retrieve the declared type "
+                               "of the node '{}'".format(self.node.name))
+
+        return ffi.string(t.name)
+
+    @typeName.setter
+    def typeName(self, type_name):
+        """Override the node's type to type_name, found using _getType.
+        The new type must resolve to the same basictype.
+
+        :param type_name: string name of the type.
+        """
+        current_override = self._override_type
+
+        declared_type = _smi.smiGetNodeType(self.node)
+        declared_basetype = self.type
+
+        new_type = _getType(type_name)
+        if not new_type:
+            raise SMIException("no type named {0} in any loaded module".format(
+                type_name))
+
+        # Easiest way to find the new basetype is to set the override
+        # and ask.
+        self._override_type = new_type
+        new_basetype = self.type
+
+        if declared_basetype != new_basetype:
+            self._override_type = current_override
+            raise SMIException("override type {1} not compatible with "
+                               "basetype of {0}".format(
+                                   ffi.string(declared_type.name),
+                                   ffi.string(new_type.name)))
+
+    @typeName.deleter
+    def typeName(self):
+        """Clears the type override."""
+        self._override_type = None
+
     @property
     def fmt(self):
         """Get node format. The node format is a string to use to display
@@ -221,7 +157,10 @@ class Node(object):
             format available.
 
         """
-        t = _smi.smiGetNodeType(self.node)
+        if self._override_type:
+            t = self._override_type
+        else:
+            t = _smi.smiGetNodeType(self.node)
         tt = _smi.smiGetParentType(t)
         f = (t != ffi.NULL and t.format != ffi.NULL and ffi.string(t.format) or
              tt != ffi.NULL and tt.format != ffi.NULL and
@@ -459,18 +398,58 @@ class Column(Node):
         return t
 
 
+_lastError = None
+
+
+@ffi.callback("void(char *, int, int, char *, char*)")
+def _logError(path, line, severity, msg, tag):
+    global _lastError
+    if path != ffi.NULL and msg != ffi.NULL:
+        _lastError = "{0}:{1}: {2}".format(ffi.string(path), line,
+                                           ffi.string(msg))
+    else:
+        _lastError = None
+
+
 def reset():
     """Reset libsmi to its initial state."""
     _smi.smiExit()
     if _smi.smiInit(b"snimpy") < 0:
             raise SMIException("unable to init libsmi")
-    _smi.smiSetErrorLevel(0)
+    _smi.smiSetErrorLevel(1)
+    _smi.smiSetErrorHandler(_logError)
     try:
         _smi.smiSetFlags(_smi.SMI_FLAG_ERRORS | _smi.SMI_FLAG_RECURSIVE)
     except TypeError:
         pass                    # We are being mocked
 
 
+def path(path=None):
+    """Set or get a search path to libsmi.
+
+    When no path is provided, return the current path,
+    unmodified. Otherwise, set the path to the specified value.
+
+    :param path: The string to be used to change the search path or
+                 `None`
+
+    """
+    if path is None:
+        # Get the path
+        path = _smi.smiGetPath()
+        if path == ffi.NULL:
+            raise SMIException("unable to get current libsmi path")
+        result = ffi.string(path)
+        _smi.free(path)
+        return result.decode("utf8")
+
+    # Set the path
+    if not isinstance(path, bytes):
+        path = path.encode("utf8")
+    if _smi.smiSetPath(path) < 0:
+        raise SMIException("unable to set the path {0}".format(path))
+
+
 def _get_module(name):
     """Get the SMI module from its name.
 
@@ -516,6 +495,35 @@ def get(mib, name):
     return pnode(node)
 
 
+def getByOid(oid):
+    """Get a node by its OID.
+
+    :param oid: The OID as a tuple
+    :return: The requested MIB node (:class:`Node`)
+    """
+    node = _smi.smiGetNodeByOID(len(oid), oid)
+    if node == ffi.NULL:
+        raise SMIException("no node for {0}".format(
+            ".".join([str(o) for o in oid])))
+    pnode = _kind2object(node.nodekind)
+    return pnode(node)
+
+
+def _getType(type_name):
+    """Searches for a smi type through all loaded modules.
+
+    :param type_name: The name of the type to search for.
+    :return: The requested type (:class:`smi.SmiType`), if found, or None.
+    """
+    if not isinstance(type_name, bytes):
+        type_name = type_name.encode("ascii")
+    for module in _loadedModules():
+        new_type = _smi.smiGetType(module, type_name)
+        if new_type != ffi.NULL:
+            return new_type
+    return None
+
+
 def _get_kind(mib, kind):
     """Get nodes of a given kind from a MIB.
 
@@ -587,11 +595,37 @@ def load(mib):
         mib = mib.encode("ascii")
     modulename = _smi.smiLoadModule(mib)
     if modulename == ffi.NULL:
-        raise SMIException("unable to load {0}".format(mib))
+        raise SMIException("unable to find {0} (check the path)".format(mib))
     modulename = ffi.string(modulename)
     if not _get_module(modulename.decode("ascii")):
-        raise SMIException("{0} contains major SMI error "
-                           "(check with smilint -s -l1)".format(mib))
+        details = "check with smilint -s -l1"
+        if _lastError is not None:
+            details = "{0}: {1}".format(_lastError,
+                                        details)
+        raise SMIException(
+            "{0} contains major SMI error ({1})".format(mib, details))
     return modulename
 
+
+def _loadedModules():
+    """Generates the list of loaded modules.
+
+    :yield: The :class:`smi.SmiModule` of all currently loaded modules.
+    """
+    module = _smi.smiGetFirstModule()
+    while module != ffi.NULL:
+        yield module
+
+        module = _smi.smiGetNextModule(module)
+
+
+def loadedMibNames():
+    """Generates the list of loaded MIB names.
+
+    :yield: The names of all currently loaded MIBs.
+    """
+    for module in _loadedModules():
+        yield ffi.string(module.name).decode('utf-8')
+
+
 reset()
